@make(ais)
@aiInfo(topic=<Model numbers of a real subtype only affect attribute values>,
number=[AI-00571/07],
classcode=[CR],
statuscode=[WA],
mainref=[03.05.07@\(15)],
date=[93-06-24])
@aiHeading()
@begin(history)
@begin(toplines)
@mainref()
@standard[03.05.09@\(14)]
@standard[03.05.08@\(03)]
@standard[03.05.08@\(16)]
@standard[03.05.10@\(03)]
@standard[03.05.10@\(15)]
@standard[04.05.07@\(00)]
!class correction 92-12-15
@cbon()!status approved by WG9 93-06-18@cboff()
@end(toplines)
!status ARG-approved 92-01-19 (4-0-0) (pending editorial review)
!status work-item 92-12-15 (returned by Chair because of major changes)
!status ARG-approved 89-10-25 (9-0-3) (pending editorial review; ballot)
!status work-item 88-10-07
!status received 88-07-06
@references[!references AI-00375, AI-00407, AI-00903, @CommentNum[83-00973]83-00973, @CommentNum[83-01050]83-01050]
@topic{!topic Model numbers of a real subtype only affect attribute values}
@end(history)

@subheading[!summary 93-01-27]

This AI reverses part of the recommendation in AI-00407, namely, the part
that allows real values to be stored with less accuracy than that of a base
type.

@subheading[!question 92-12-15]

AI-00407 says (in its summary):
@begin(quotation)
When assigning a fixed or floating point value to a variable, the
stored value need only be represented as a model number of the
variable's subtype.  Furthermore, if no exception is raised by the
assignment, the stored value belongs to the subtype of the variable.
@end(quotation)
The purpose of this interpretation was to allow real values to be stored with
less accuracy, since this was understood to be the original intent of the
language design.

As a consequence of this part of AI-00407, assignments have the potentially
surprising property that they can cause an @i[implicit] loss of accuracy.
Consider the following example, which is essentially identical to the example
in AI-00407.
@begin(example)
@key[declare]
    @key[type] T @key[is] @key[digits] 5;
    @key[subtype] ST @key[is] T @key[digits] 3 @key[range] 12345.0 .. 15099.0;
    X : ST;
@key[begin]
    X := 12345.0;
@end(example)
AI-00407 permits an implementation to represent X using a shorter storage
representation than is possible for variables of type T.  According to the
AI, an implementation is only obliged to retain 3 decimal digits (more
precisely, 11 binary digits) in the mantissa of values assigned to X, in
contrast to the 5 decimal (18 binary) digits required for variables of type
T.  The user must not assume that X can contain values with a granularity
finer than that of the model numbers of ST, because an implementation that
discards or alters mantissa digits (and thus widens the model interval) of a
value being assigned to X, after the equivalent of the 11th bit, is valid
according to the AI.  Indeed, in the assignment to X, the value 12345.0,
which is not a model number of ST, may be perturbed to 12344.0, an endpoint
of the model interval of ST containing 12345.0.  The AI points out that the
perturbed value, 12344.0, must be the subject of the constraint check and
would, in the present case, raise CONSTRAINT_@|ERROR.

The use of explicit conversions to reduce accuracy is much more consistent
with the goals of the Ada design.  For example, consider the use of explicit
conversions in the following example:
@begin(example)
@key[declare]
    A    : LONG_@|FLOAT;
    B, C : FLOAT;
@key[begin]
    B := FLOAT (A + LONG_@|FLOAT(B) @multiply() LONG_@|FLOAT(C));
@key[end];
@end(example)
The possible loss of accuracy is explicit in the conversion of the
sum to type FLOAT.  On the other hand, with the ruling of AI-00407, the same
effect could be obtained implicitly:
@begin(example)
@key[declare]
    @key[subtype] SINGLE @key[is] LONG_@|FLOAT @key[digits] FLOAT'@|DIGITS
                        @key[range] FLOAT'@|FIRST .. FLOAT'@|LAST;
    A    : LONG_@|FLOAT;
    B, C : SINGLE;
@key[begin]
    B := A + B@multiply()C;
@key[end];
@end(example)
AI-00407 allows an implementation to store values of subtype SINGLE with
reduced accuracy since the range of the subtype allows such values to be
stored.  Although the expression in the second example may be more convenient
to write, it is much more subject to misinterpretation, especially in a
strongly typed language like Ada.  Shouldn't assignment statements be
prohibited from storing values with reduced accuracy?  Won't programmers be
surprised by such an effect?

Does a fixed-point constraint or a floating-point constraint serve any useful
purpose in a subtype declaration, if it does not permit a shorter
representation?

@subheading[!recommendation 93-01-27]

Consider the following part of the recommendation in AI-00407:
@begin(quotation)
If the subtype of the variable in an assignment statement has less
accuracy than the type of the expression, the model interval of the
expression is widened to the smallest containing model interval of the
subtype.  Any value in the widened model interval can become the value of
the variable, but CONSTRAINT_@|ERROR is raised if the value to be stored
does not belong to the variable's subtype.
@end(quotation)
This part of the recommendation is replaced with:
@begin(quotation)
When assigning a fixed or a floating point value to a variable, the
stored value must be represented with at least the accuracy of the base
type.
@end(quotation)
@subheading[!discussion 93-01-27]

The rest of the recommendation of AI-00407 reads as follows (and is not
superseded):
@begin(quotation)
If a real subtype is used as the type mark in a membership test,
qualification, or explicit conversion, the corresponding operation is
performed with the accuracy of the base type and the range of the subtype.

For a real subtype, the value of the attribute FIRST or LAST is
represented with at least the accuracy of the base type.  The values of
other attributes of a real subtype are given exactly.
@end(quotation)
AI-00407 allows fixed point and floating point values to be stored with
reduced accuracy if a variable's subtype has a reduced accuracy constraint,
i.e., a floating accuracy definition or fixed accuracy definition that
reduces the digits or increases the delta of the base type.  Further
consideration of this ruling has shown that it is harmful to the use of Ada.
The main purpose of the current commentary is to rescind this ruling of
AI-00407 and to make clearer the effect of subtype declarations with reduced
accuracy constraints.

One reason for approving AI-00407 was to facilitate use of the IEEE floating
point standard, which mandates support for mixed precision floating point
arithmetic.  In order to allow support for arithmetic on values with a reduced
mantissa length under AI-00407, however, one must declare a reduced precision
subtype with a range constraint; a reduced precision subtype cannot be
obtained simply by declaring a subtype with an accuracy constraint:
@begin(example)
@key[subtype] SFLOAT @key[is] LONG_@|FLOAT @key[digits] 5;
X : SFLOAT := SFLOAT'@|LAST;             -- same as LONG_@|FLOAT'@|LAST
@end(example)
AI-00174 states that SFLOAT'@|LAST equals LONG_@|FLOAT'@|LAST, since no range
constraint is specified for SFLOAT.  Moreover, AI-00174 states that
SFLOAT'@|LAST is the largest storable value of the subtype.  SFLOAT's stored
values cannot have a reduced mantissa length, since any reduction in length
will make it impossible to store the value of SFLOAT'@|LAST.  Therefore, to
allow the implementation to store values in a reduced precision format under
AI-00407, one must declare a suitable range constraint:
@begin(example)
@key[type] FLOAT5 @key[is] @key[digits] 5;
S_@|FIRST : @key[constant] LONG_@|FLOAT := LONG_@|FLOAT(FLOAT5'FIRST);
S_@|LAST  : @key[constant] LONG_@|FLOAT := LONG_@|FLOAT(FLOAT5'LAST);
@key[subtype] S5_@|FLOAT @key[is] LONG_@|FLOAT @key[digits] 5 @key[range] S_@|FIRST .. S_@|LAST;
@end(example)
If LONG_@|FLOAT indeed has a larger mantissa than FLOAT5's base type, and
LONG_@|FLOAT's exponent range is at least as big, then the conversion for
FLOAT5'FIRST and FLOAT5'LAST to LONG_@|FLOAT will not change the value, and
S5_@|FLOAT will, in practice, be declared to have exactly the representable
range of a reduced precision subtype.  Under AI-00407, the implementation
would then be able to implicitly store LONG_@|FLOAT values with less precision
when assigning values to a variable of subtype S5_@|FLOAT.  So although AI-00407
allows a form of @dq"mixed mode" arithmetic to be supported in Ada, the needed
reduced precision subtypes cannot be obtained quite as conveniently as one
might think.

Although mixed precision arithmetic is common in other programming languages,
implicit changes in values stored in variables having the same base type (due
to an implicit change in the accuracy of stored values) is inconsistent with
the general type-safe spirit of the Ada design.  Explicit conversions are the
preferred method of converting a value to a less-precise stored representation.

In short, the capability provided by AI-00407 increases the danger of subtle
programmer errors.  Moreover, since no implementations are known to support
AI-00407 for floating point types, there is no harm in reversing its
conclusion.

Similar declarations exist for fixed point types:
@begin(example)
@key[type] FIXED @key[is] @key[delta] 2.0@exp()(@t[-]15) @key[range] @t[-]1.0 .. 1.0;
@key[subtype] SFIXED @key[is] FIXED @key[delta] 2.0@exp()(@t[-]7);
@end(example)
If an implementation supports the storage of reduced accuracy values, then
values of subtype SFIXED can be stored in 8 bits, while values of type FIXED
require 16 bits.  A reduced accuracy fixed point representations might be of
some value in a record type, where it might make sense to store values of
subtype SFIXED in only 8 bits.  However, the arguments against implicit loss
of accuracy in an assignment statement still apply to the fixed point case.
It is still better for programmers to declare different fixed point types if
they want values to be represented with different accuracies.  No
implementations are known to use fewer bits of precision to represent stored
values of a subtype like SFIXED.

In short, AI-00407's support for storing values with reduced accuracy is not
well motivated and should be reversed.

Given the inability to store values of a real type with reduced accuracy, and
given that all operations of a real type are performed with the accuracy of
the base type (for reasons explained in AI-00407), is there any remaining
motivation for the notion that real subtypes declared with accuracy
constraints have a different set of model numbers than the base type?  The
only remaining effect of such model numbers is in determining the values of
certain attributes.  For floating point subtypes, these attributes are DIGITS,
MANTISSA, EPSILON, EMAX, SMALL, and LARGE.  Of these attributes, only DIGITS
has a significant use in the language@dash()it determines the value of
DEFAULT_@|AFT in the FLOAT_@|IO package.  Since the model numbers of a subtype
have no effect on a program except to define the values of these attributes,
it would probably be less confusing if Ada did not bother to redefine the
model numbers associated with such subtypes, simply assuming that the model
numbers of a floating point subtype are those of its base type and that the
effect of a floating accuracy definition is to define the value of the
subtype's DIGITS attribute.  Although such a change is attractive for
improving the simplicity of the language (by removing the almost-useless
concept of model numbers for real subtypes), such a change would only cause
work for implementors with no real benefits for existing users.

Similarly, of the attributes associated with the model numbers of a fixed
point subtype, namely, DELTA, MANTISSA, SMALL, and LARGE, only DELTA has a
significant use in defining the value of DEFAULT_@|AFT in the FIXED_@|IO package.
So it would also seem reasonable to define the model numbers of a fixed point
subtype to be those of its base type, with the effect of a fixed accuracy
definition simply being to define the value of the subtype's DELTA attribute,
thereby indirectly defining the value of the subtype's AFT attribute. (The
value of the FORE attribute is determined by the subtype's range).  This
would also simplify the computation of the value of the MANTISSA attribute,
which must sometimes now be computed at run-time (see AI-00145).  But as
in the floating point case, such changes would not be well motivated by
ambiguities in the wording of the standard or by benefits to existing users,
so the changes cannot be recommended.

In short, although the notion of model numbers for a real subtype has almost
no effect on the semantics of programs, there is insufficient basis for
removing the concept entirely.

